@@ -1,7 +1,9 @@
 #ifndef _checker_hpp_INCLUDED
 #define _checker_hpp_INCLUDED
 
-/*------------------------------------------------------------------------*/
+#include "tracer.hpp" // Alphabetically after 'checker'.
+
+#include <cstdint>
 
 namespace CaDiCaL {
 
@@ -45,7 +47,7 @@ typedef vector<CheckerWatch> CheckerWatcher;
 
 /*------------------------------------------------------------------------*/
 
-class Checker {
+class Checker : public StatTracer {
 
   Internal *internal;
 
@@ -147,11 +149,20 @@ public:
   Checker (Internal *);
   ~Checker ();
 
-  void add_original_clause (uint64_t, const vector<int> &);
-  void add_derived_clause (uint64_t, const vector<int> &);
-  void delete_clause (uint64_t, const vector<int> &);
+  void connect_internal (Internal *i) override;
 
-  void print_stats ();
+  void add_original_clause (uint64_t, bool, const vector<int> &,
+                            bool = false) override;
+  void add_derived_clause (uint64_t, bool, const vector<int> &,
+                           const vector<uint64_t> &) override;
+  void delete_clause (uint64_t, bool, const vector<int> &) override;
+
+  void finalize_clause (uint64_t, const vector<int> &) override {} // skip
+  void report_status (int, uint64_t) override {}                   // skip
+  void begin_proof (uint64_t) override {}                          // skip
+  void add_assumption_clause (uint64_t, const vector<int> &,
+                              const vector<uint64_t> &) override;
+  void print_stats () override;
   void dump (); // for debugging purposes only
 };
 
