@@ -31,16 +31,13 @@ namespace CaDiCaL {
 // Private constructor.
 
 File::File (Internal *i, bool w, int c, int p, FILE *f, const char *n)
-    :
-#ifndef QUIET
-      internal (i),
-#endif
+    : internal (i),
 #if !defined(QUIET) || !defined(NDEBUG)
       writing (w),
 #endif
-      close_file (c), child_pid (p), file (f), _name (n), _lineno (1),
-      _bytes (0) {
-  (void) i, (void) w;
+      close_file (c), child_pid (p), file (f), _name (strdup (n)),
+      _lineno (1), _bytes (0) {
+  (void) w;
   assert (f), assert (n);
 }
 
@@ -98,6 +95,14 @@ bool File::writable (const char *path) {
   return !res;
 }
 
+bool File::piping () {
+  struct stat stat;
+  int fd = fileno (file);
+  if (fstat (fd, &stat))
+    return true;
+  return S_ISFIFO (stat.st_mode);
+}
+
 // These are signatures for supported compressed file types.  In 2018 the
 // SAT Competition was running on StarExec and used internally 'bzip2'
 // compressed files, but gave them uncompressed to the solver using exactly
@@ -256,6 +261,9 @@ FILE *File::write_pipe (Internal *internal, const char *command,
                         const char *path, int &child_pid) {
   assert (command[0] && command[0] != ' ');
   MSG ("writing through command '%s' to '%s'", command, path);
+#ifdef QUIET
+  (void) internal;
+#endif
   std::vector<char *> args;
   split_str (command, args);
   assert (!args.empty ());
@@ -292,6 +300,9 @@ FILE *File::write_pipe (Internal *internal, const char *command,
   if (absolute_command_path)
     delete[] absolute_command_path;
   delete_str_vector (args);
+#ifdef QUIET
+  (void) internal;
+#endif
   return res;
 }
 
@@ -364,22 +375,32 @@ File *File::write (Internal *internal, const char *path) {
   return new File (internal, true, close_output, child_pid, file, path);
 }
 
-void File::close () {
+void File::close (bool print) {
   assert (file);
+#ifndef QUIET
+  if (internal->opts.quiet)
+    print = false;
+  else if (internal->opts.verbose > 0)
+    print = true;
+#endif
   if (close_file == 0) {
-    MSG ("disconnecting from '%s'", name ());
+    if (print)
+      MSG ("disconnecting from '%s'", name ());
   }
   if (close_file == 1) {
-    MSG ("closing file '%s'", name ());
+    if (print)
+      MSG ("closing file '%s'", name ());
     fclose (file);
   }
   if (close_file == 2) {
-    MSG ("closing input pipe to read '%s'", name ());
+    if (print)
+      MSG ("closing input pipe to read '%s'", name ());
     pclose (file);
   }
 #ifndef _WIN32
   if (close_file == 3) {
-    MSG ("closing output pipe to write '%s'", name ());
+    if (print)
+      MSG ("closing output pipe to write '%s'", name ());
     fclose (file);
     waitpid (child_pid, 0, 0);
   }
@@ -389,7 +410,7 @@ void File::close () {
   // TODO what about error checking for 'fclose', 'pclose' or 'waitpid'?
 
 #ifndef QUIET
-  if (internal->opts.verbose > 1) {
+  if (print) {
     if (writing) {
       uint64_t written_bytes = bytes ();
       double written_mb = written_bytes / (double) (1 << 20);
@@ -431,6 +452,7 @@ void File::flush () {
 File::~File () {
   if (file)
     close ();
+  free (_name);
 }
 
 } // namespace CaDiCaL
