@@ -59,18 +59,22 @@ namespace CaDiCaL {
 
 bool Internal::subsuming () {
 
-  if (!opts.subsume && !opts.vivify) return false;
-  if (!preprocessing && !opts.inprocessing) return false;
-  if (preprocessing) assert (lim.preprocessing);
+  if (!opts.subsume && !opts.vivify)
+    return false;
+  if (!preprocessing && !opts.inprocessing)
+    return false;
+  if (preprocessing)
+    assert (lim.preprocessing);
 
   // Only perform global subsumption checking immediately after a clause
   // reduction happened where the overall allocated memory is small and we
   // got a limit on the number of kept clause in terms of size and glue.
   //
-  if (opts.reduce &&
-      stats.conflicts != last.reduce.conflicts) return false;
+  if (opts.reduce && stats.conflicts != last.reduce.conflicts)
+    return false;
 
-  if (stats.conflicts < lim.subsume) return false;
+  if (stats.conflicts < lim.subsume)
+    return false;
 
   return true;
 }
@@ -89,8 +93,7 @@ bool Internal::subsuming () {
 // strengthened and as a result the negation of the literal which can be
 // removed is returned.
 
-inline int Internal::subsume_check (Clause * subsuming, Clause * subsumed)
-{
+inline int Internal::subsume_check (Clause *subsuming, Clause *subsumed) {
 #ifdef NDEBUG
   (void) subsumed;
 #endif
@@ -103,7 +106,8 @@ inline int Internal::subsume_check (Clause * subsuming, Clause * subsumed)
   assert (subsuming->size <= subsumed->size);
 
   stats.subchecks++;
-  if (subsuming->size == 2) stats.subchecks2++;
+  if (subsuming->size == 2)
+    stats.subchecks2++;
 
   int flipped = 0, prev = 0;
   bool failed = false;
@@ -113,38 +117,53 @@ inline int Internal::subsume_check (Clause * subsuming, Clause * subsumed)
     *i = prev;
     prev = lit;
     const int tmp = marked (lit);
-    if (!tmp) failed = true;
-    else if (tmp > 0) continue;
-    else if (flipped) failed = true;
-    else flipped = lit;
+    if (!tmp)
+      failed = true;
+    else if (tmp > 0)
+      continue;
+    else if (flipped)
+      failed = true;
+    else
+      flipped = lit;
   }
   assert (prev);
   assert (!subsuming->literals[0]);
   subsuming->literals[0] = prev;
-  if (failed) return 0;
+  if (failed)
+    return 0;
 
-  if (!flipped) return INT_MIN;                   // subsumed!!
-  else if (!opts.subsumestr) return 0;
-  else return flipped;                            // strengthen!!
+  if (!flipped)
+    return INT_MIN; // subsumed!!
+  else if (!opts.subsumestr)
+    return 0;
+  else
+    return flipped; // strengthen!!
 }
 
 /*------------------------------------------------------------------------*/
 
 // Candidate clause 'subsumed' is subsumed by 'subsuming'.
 
-inline void
-Internal::subsume_clause (Clause * subsuming, Clause * subsumed) {
+inline void Internal::subsume_clause (Clause *subsuming, Clause *subsumed) {
   stats.subsumed++;
   assert (subsuming->size <= subsumed->size);
   LOG (subsumed, "subsumed");
-  if (subsumed->redundant) stats.subred++; else stats.subirr++;
-  mark_garbage (subsumed);
-  if (subsumed->redundant || !subsuming->redundant) return;
+  if (subsumed->redundant)
+    stats.subred++;
+  else
+    stats.subirr++;
+  if (subsumed->redundant || !subsuming->redundant) {
+    mark_garbage (subsumed);
+    return;
+  }
   LOG ("turning redundant subsuming clause into irredundant clause");
   subsuming->redundant = false;
+  if (proof)
+    proof->strengthen (subsuming->id);
+  mark_garbage (subsumed);
   stats.current.irredundant++;
   stats.added.irredundant++;
-  stats.irrbytes += subsuming->bytes ();
+  stats.irrlits += subsuming->size;
   assert (stats.current.redundant > 0);
   stats.current.redundant--;
   assert (stats.added.redundant > 0);
@@ -156,16 +175,19 @@ Internal::subsume_clause (Clause * subsuming, Clause * subsumed) {
 
 // Candidate clause 'c' is strengthened by removing 'lit'.
 
-void Internal::strengthen_clause (Clause * c, int lit) {
+void Internal::strengthen_clause (Clause *c, int lit) {
   stats.strengthened++;
   assert (c->size > 2);
   LOG (c, "removing %d in", lit);
-  if (proof) proof->strengthen_clause (c, lit);
-  if (!c->redundant) mark_removed (lit);
+  if (proof) {
+    LOG (lrat_chain, "strengthening clause with chain");
+    proof->strengthen_clause (c, lit, lrat_chain);
+  }
+  if (!c->redundant)
+    mark_removed (lit);
   auto new_end = remove (c->begin (), c->end (), lit);
   assert (new_end + 1 == c->end ()), (void) new_end;
   (void) shrink_clause (c, c->size - 1);
-  c->used = true;
   LOG (c, "strengthened");
   external->check_shrunken_clause (c);
 }
@@ -178,21 +200,21 @@ void Internal::strengthen_clause (Clause * c, int lit) {
 // strengthened the result is negative.  Otherwise the candidate clause
 // can not be subsumed nor strengthened and zero is returned.
 
-inline int
-Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
+inline int Internal::try_to_subsume_clause (Clause *c,
+                                            vector<Clause *> &shrunken) {
 
   stats.subtried++;
   assert (!level);
   LOG (c, "trying to subsume");
 
-  mark (c);     // signed!
+  mark (c); // signed!
 
   Clause dummy; // Communicate binary subsuming clause.
 
-  Clause * d = 0;
+  Clause *d = 0;
   int flipped = 0;
 
-  for (const auto & lit : *c) {
+  for (const auto &lit : *c) {
 
     // Only clauses which have a variable which has recently been added are
     // checked for being subsumed.  The idea is that all these newly added
@@ -200,7 +222,8 @@ Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
     // need to check occurrences of these variables.  The occurrence lists
     // of other literal do not have to be checked.
     //
-    if (!flags (lit).subsume) continue;
+    if (!flags (lit).subsume)
+      continue;
 
     for (int sign = -1; !d && sign <= 1; sign += 2) {
 
@@ -217,27 +240,34 @@ Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
       // removed in 'c', otherwise to 'INT_MIN' which is a non-valid
       // literal.
       //
-      for (const auto & other : bins (sign*lit) ) {
+      for (const auto &bin : bins (sign * lit)) {
+        const auto &other = bin.lit;
         const int tmp = marked (other);
-        if (!tmp) continue;
-        if (tmp < 0 && sign < 0) continue;
+        if (!tmp)
+          continue;
+        if (tmp < 0 && sign < 0)
+          continue;
         if (tmp < 0) {
-          if (sign < 0) continue;               // tautological resolvent
+          if (sign < 0)
+            continue; // tautological resolvent
           dummy.literals[0] = lit;
           dummy.literals[1] = other;
           flipped = other;
         } else {
-          dummy.literals[0] = sign*lit;
+          dummy.literals[0] = sign * lit;
           dummy.literals[1] = other;
           flipped = (sign < 0) ? -lit : INT_MIN;
         }
+        dummy.moved = false;
         dummy.redundant = false;
         dummy.size = 2;
+        dummy.id = bin.id;
         d = &dummy;
         break;
       }
 
-      if (d) break;
+      if (d)
+        break;
 
       // In this second loop we check for larger than binary clauses to
       // subsume or strengthen the candidate clause.   This is more costly,
@@ -245,18 +275,21 @@ Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
       // as above for communicating 'subsumption' or 'strengthening' to the
       // code after the loop is used.
       //
-      const Occs & os = occs (sign * lit);
-      for (const auto & e : os) {
-        assert (!e->garbage);                   // sanity check
-        if (e->garbage) continue;               // defensive: not needed
+      const Occs &os = occs (sign * lit);
+      for (const auto &e : os) {
+        assert (!e->garbage); // sanity check
+        if (e->garbage)
+          continue; // defensive: not needed
         flipped = subsume_check (e, c);
-        if (!flipped) continue;
-        d = e;                                  // leave also outer loop
+        if (!flipped)
+          continue;
+        d = e; // leave also outer loop
         break;
       }
     }
 
-    if (d) break;
+    if (d)
+      break;
   }
 
   unmark (c);
@@ -269,7 +302,13 @@ Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
 
   if (flipped) {
     LOG (d, "strengthening");
+    if (lrat) {
+      assert (lrat_chain.empty ());
+      lrat_chain.push_back (c->id);
+      lrat_chain.push_back (d->id);
+    }
     strengthen_clause (c, -flipped);
+    lrat_chain.clear ();
     assert (likely_to_be_kept_clause (c));
     shrunken.push_back (c);
     return -1;
@@ -286,28 +325,32 @@ Internal::try_to_subsume_clause (Clause * c, vector<Clause *> & shrunken) {
 
 struct ClauseSize {
   size_t size;
-  Clause * clause;
-  ClauseSize (int s, Clause * c) : size (s), clause (c) { }
-  ClauseSize () { }
+  Clause *clause;
+  ClauseSize (int s, Clause *c) : size (s), clause (c) {}
+  ClauseSize () {}
 };
 
 struct smaller_clause_size_rank {
   typedef size_t Type;
-  Type operator () (const ClauseSize & a) { return a.size; }
+  Type operator() (const ClauseSize &a) { return a.size; }
 };
 
 /*------------------------------------------------------------------------*/
 
 struct subsume_less_noccs {
-  Internal * internal;
-  subsume_less_noccs (Internal * i) : internal (i) { }
-  bool operator () (int a, int b) {
+  Internal *internal;
+  subsume_less_noccs (Internal *i) : internal (i) {}
+  bool operator() (int a, int b) {
     const signed char u = internal->val (a), v = internal->val (b);
-    if (!u && v) return true;
-    if (u && !v) return false;
-    const int64_t m = internal->noccs (a), n = internal ->noccs (b);
-    if (m < n) return true;
-    if (m > n) return false;
+    if (!u && v)
+      return true;
+    if (u && !v)
+      return false;
+    const int64_t m = internal->noccs (a), n = internal->noccs (b);
+    if (m < n)
+      return true;
+    if (m > n)
+      return false;
     return abs (a) < abs (b);
   }
 };
@@ -324,10 +367,14 @@ struct subsume_less_noccs {
 
 bool Internal::subsume_round () {
 
-  if (!opts.subsume) return false;
-  if (unsat) return false;
-  if (terminated_asynchronously ()) return false;
-  if (!stats.current.redundant && !stats.current.irredundant) return false;
+  if (!opts.subsume)
+    return false;
+  if (unsat)
+    return false;
+  if (terminated_asynchronously ())
+    return false;
+  if (!stats.current.redundant && !stats.current.irredundant)
+    return false;
 
   START_SIMPLIFIER (subsume, SUBSUME);
   stats.subsumerounds++;
@@ -336,17 +383,19 @@ bool Internal::subsume_round () {
   if (opts.subsumelimited) {
     int64_t delta = stats.propagations.search;
     delta *= 1e-3 * opts.subsumereleff;
-    if (delta < opts.subsumemineff) delta = opts.subsumemineff;
-    if (delta > opts.subsumemaxeff) delta = opts.subsumemaxeff;
+    if (delta < opts.subsumemineff)
+      delta = opts.subsumemineff;
+    if (delta > opts.subsumemaxeff)
+      delta = opts.subsumemaxeff;
     delta = max (delta, (int64_t) 2l * active ());
 
     PHASE ("subsume-round", stats.subsumerounds,
-      "limit of %" PRId64 " subsumption checks", delta);
+           "limit of %" PRId64 " subsumption checks", delta);
 
     check_limit = stats.subchecks + delta;
   } else {
     PHASE ("subsume-round", stats.subsumerounds,
-      "unlimited subsumption checks");
+           "unlimited subsumption checks");
     check_limit = LONG_MAX;
   }
 
@@ -365,15 +414,20 @@ bool Internal::subsume_round () {
 
   for (auto c : clauses) {
 
-    if (c->garbage) continue;
-    if (c->size > opts.subsumeclslim) continue;
-    if (!likely_to_be_kept_clause (c)) continue;
+    if (c->garbage)
+      continue;
+    if (c->size > opts.subsumeclslim)
+      continue;
+    if (!likely_to_be_kept_clause (c))
+      continue;
 
     bool fixed = false;
     int subsume = 0;
-    for (const auto & lit : *c)
-      if (val (lit)) fixed = true;
-      else if (flags (lit).subsume) subsume++;
+    for (const auto &lit : *c)
+      if (val (lit))
+        fixed = true;
+      else if (flags (lit).subsume)
+        subsume++;
 
     // If the clause contains a root level assigned (fixed) literal we will
     // not work on it.  This simplifies the code substantially since we do
@@ -393,9 +447,10 @@ bool Internal::subsume_round () {
       continue;
     }
 
-    if (c->subsume) left_over_from_last_subsumption_round++;
+    if (c->subsume)
+      left_over_from_last_subsumption_round++;
     schedule.push_back (ClauseSize (c->size, c));
-    for (const auto & lit : *c)
+    for (const auto &lit : *c)
       noccs (lit)++;
   }
   shrink_vector (schedule);
@@ -413,8 +468,8 @@ bool Internal::subsume_round () {
   int64_t scheduled = schedule.size ();
   int64_t total = stats.current.irredundant + stats.current.redundant;
   PHASE ("subsume-round", stats.subsumerounds,
-    "scheduled %" PRId64 " clauses %.0f%% out of %" PRId64 " clauses",
-    scheduled, percent (scheduled, total), total);
+         "scheduled %" PRId64 " clauses %.0f%% out of %" PRId64 " clauses",
+         scheduled, percent (scheduled, total), total);
 #endif
 
   // Now go over the scheduled clauses in the order of increasing size and
@@ -430,12 +485,14 @@ bool Internal::subsume_round () {
   init_occs ();
   init_bins ();
 
-  for (const auto & s : schedule) {
+  for (const auto &s : schedule) {
 
-    if (terminated_asynchronously ()) break;
-    if (stats.subchecks >= check_limit) break;
+    if (terminated_asynchronously ())
+      break;
+    if (stats.subchecks >= check_limit)
+      break;
 
-    Clause * c = s.clause;
+    Clause *c = s.clause;
     assert (!c->garbage);
 
     checked++;
@@ -451,8 +508,12 @@ bool Internal::subsume_round () {
     if (c->size > 2 && c->subsume) {
       c->subsume = false;
       const int tmp = try_to_subsume_clause (c, shrunken);
-      if (tmp > 0) { subsumed++; continue; }
-      if (tmp < 0) strengthened++;
+      if (tmp > 0) {
+        subsumed++;
+        continue;
+      }
+      if (tmp < 0)
+        strengthened++;
     }
 
     // If not subsumed connect smallest occurring literal, where occurring
@@ -468,13 +529,16 @@ bool Internal::subsume_round () {
     bool subsume = true;
     bool binary = (c->size == 2 && !c->redundant);
 
-    for (const auto & lit : *c) {
+    for (const auto &lit : *c) {
 
-      if (!flags (lit).subsume) subsume = false;
+      if (!flags (lit).subsume)
+        subsume = false;
       const size_t size = binary ? bins (lit).size () : occs (lit).size ();
-      if (minlit && minsize <= size) continue;
+      if (minlit && minsize <= size)
+        continue;
       const int64_t tmp = noccs (lit);
-      if (minlit && minsize == size && tmp <= minoccs) continue;
+      if (minlit && minsize == size && tmp <= minoccs)
+        continue;
       minlit = lit, minsize = size, minoccs = tmp;
     }
 
@@ -483,16 +547,19 @@ bool Internal::subsume_round () {
     // can not serve to strengthen or subsume another clause, since all
     // shrunken or added clauses mark all their variables as 'subsume'.
     //
-    if (!subsume) continue;
+    if (!subsume)
+      continue;
 
     if (!binary) {
 
       // If smallest occurring literal occurs too often do not connect.
       //
-      if (minsize > (size_t) opts.subsumeocclim) continue;
+      if (minsize > (size_t) opts.subsumeocclim)
+        continue;
 
-      LOG (c, "watching %d with %zd current and total %" PRId64 " occurrences",
-        minlit, minsize, minoccs);
+      LOG (c,
+           "watching %d with %zd current and total %" PRId64 " occurrences",
+           minlit, minsize, minoccs);
 
       occs (minlit).push_back (c);
 
@@ -511,34 +578,37 @@ bool Internal::subsume_round () {
 
       // If smallest occurring literal occurs too often do not connect.
       //
-      if (minsize > (size_t) opts.subsumebinlim) continue;
+      if (minsize > (size_t) opts.subsumebinlim)
+        continue;
 
       LOG (c,
-        "watching %d with %zd current binary and total %" PRId64 " occurrences",
-        minlit, minsize, minoccs);
+           "watching %d with %zd current binary and total %" PRId64
+           " occurrences",
+           minlit, minsize, minoccs);
 
       const int minlit_pos = (c->literals[1] == minlit);
       const int other = c->literals[!minlit_pos];
-      bins (minlit).push_back (other);
+      bins (minlit).push_back (Bin{other, c->id});
     }
   }
 
   PHASE ("subsume-round", stats.subsumerounds,
-    "subsumed %" PRId64 " and strengthened %" PRId64
-    " out of %" PRId64 " clauses %.0f%%",
-    subsumed, strengthened, scheduled,
-    percent (subsumed + strengthened, scheduled));
+         "subsumed %" PRId64 " and strengthened %" PRId64 " out of %" PRId64
+         " clauses %.0f%%",
+         subsumed, strengthened, scheduled,
+         percent (subsumed + strengthened, scheduled));
 
   const int64_t remain = schedule.size () - checked;
   const bool completed = !remain;
 
   if (completed)
     PHASE ("subsume-round", stats.subsumerounds,
-      "checked all %" PRId64 " scheduled clauses", checked);
+           "checked all %" PRId64 " scheduled clauses", checked);
   else
     PHASE ("subsume-round", stats.subsumerounds,
-      "checked %" PRId64 " clauses %.0f%% of scheduled (%" PRId64 " remain)",
-      checked, percent (checked, scheduled), remain);
+           "checked %" PRId64 " clauses %.0f%% of scheduled (%" PRId64
+           " remain)",
+           checked, percent (checked, scheduled), remain);
 
   // Release occurrence lists and schedule.
   //
@@ -553,7 +623,7 @@ bool Internal::subsume_round () {
   if (completed)
     reset_subsume_bits ();
 
-  for (const auto & c : shrunken)
+  for (const auto &c : shrunken)
     mark_added (c);
   erase_vector (shrunken);
 
@@ -573,7 +643,8 @@ void Internal::subsume (bool update_limits) {
   if (!stats.current.redundant && !stats.current.irredundant)
     goto UPDATE_LIMITS;
 
-  if (unsat) return;
+  if (unsat)
+    return;
 
   backtrack ();
   if (!propagate ()) {
@@ -594,19 +665,22 @@ void Internal::subsume (bool update_limits) {
 
   // Schedule 'vivification' in 'subsume' as well as 'transitive reduction'.
   //
-  if (opts.vivify) vivify ();
-  if (opts.transred) transred ();
+  if (opts.vivify)
+    vivify ();
+  if (opts.transred)
+    transred ();
 
 UPDATE_LIMITS:
 
-  if (!update_limits) return;
+  if (!update_limits)
+    return;
 
   int64_t delta = scale (opts.subsumeint * (stats.subsumephases + 1));
   lim.subsume = stats.conflicts + delta;
 
   PHASE ("subsume-phase", stats.subsumephases,
-    "new subsume limit %" PRId64 " after %" PRId64 " conflicts",
-    lim.subsume, delta);
+         "new subsume limit %" PRId64 " after %" PRId64 " conflicts",
+         lim.subsume, delta);
 }
 
-}
+} // namespace CaDiCaL
